=== modified file 'lib/canonical/launchpad/daemons/tachandler.py'
--- lib/canonical/launchpad/daemons/tachandler.py	2010-03-22 11:20:50 +0000
+++ lib/canonical/launchpad/daemons/tachandler.py	2010-04-05 09:22:54 +0000
@@ -70,7 +70,7 @@
         pass
 
 
-def kill_by_pidfile(pidfile_path):
+def kill_by_pidfile(pidfile_path, poll_interval=0.1, num_polls=50):
     """Kill a process identified by the pid stored in a file.
 
     The pid file is removed from disk.

=== modified file 'lib/canonical/testing/layers.py'
--- lib/canonical/testing/layers.py	2010-03-26 13:23:20 +0000
+++ lib/canonical/testing/layers.py	2010-04-05 09:22:54 +0000
@@ -255,9 +255,10 @@
             os.environ.get('LP_PERSISTENT_TEST_SERVICES') is not None)
         # Kill any Memcached or Librarian left running from a previous
         # test run, or from the parent test process if the current
-        # layer is being run in a subprocess.
+        # layer is being run in a subprocess. No need to be polite
+        # about killing memcached - just do it quickly.
         if not BaseLayer.persist_test_services:
-            kill_by_pidfile(MemcachedLayer.getPidFile())
+            kill_by_pidfile(MemcachedLayer.getPidFile(), num_polls=0)
             LibrarianTestSetup().tearDown()
         # Kill any database left lying around from a previous test run.
         try:
@@ -1489,6 +1490,7 @@
     @classmethod
     @profiled
     def startStory(cls):
+        MemcachedLayer.testSetUp()
         DatabaseLayer.testSetUp()
         LibrarianLayer.testSetUp()
         LaunchpadLayer.resetSessionDb()
@@ -1500,6 +1502,7 @@
         PageTestLayer.resetBetweenTests(True)
         LibrarianLayer.testTearDown()
         DatabaseLayer.testTearDown()
+        MemcachedLayer.testTearDown()
 
     @classmethod
     @profiled

=== modified file 'lib/lp/bugs/stories/bugattachments/10-add-bug-attachment.txt'
--- lib/lp/bugs/stories/bugattachments/10-add-bug-attachment.txt	2010-03-30 20:56:51 +0000
+++ lib/lp/bugs/stories/bugattachments/10-add-bug-attachment.txt	2010-04-06 06:05:25 +0000
@@ -50,6 +50,8 @@
 
   >>> '   Some information   ' in user_browser.contents
   False
+  >>> 'Some information' in user_browser.contents
+  True
 
 If no description is given it gets set to the attachment filename. It's
 also not necessary to enter a comment in order to add an attachment.

=== modified file 'lib/lp/bugs/stories/bugs/xx-remote-bug-comments.txt'
--- lib/lp/bugs/stories/bugs/xx-remote-bug-comments.txt	2009-09-02 12:33:41 +0000
+++ lib/lp/bugs/stories/bugs/xx-remote-bug-comments.txt	2010-03-30 09:00:29 +0000
@@ -93,6 +93,8 @@
     >>> flush_database_updates()
     >>> logout()
 
+    >>> from canonical.testing.layers import MemcachedLayer
+    >>> MemcachedLayer.purge()
     >>> user_browser.open('http://bugs.launchpad.dev/redfish/+bug/15')
     >>> last_bug_comment = find_tags_by_class(
     ...     user_browser.contents, 'remoteBugComment')[-1]

=== modified file 'lib/lp/bugs/templates/bugcomment-box.pt'
--- lib/lp/bugs/templates/bugcomment-box.pt	2009-12-10 14:44:24 +0000
+++ lib/lp/bugs/templates/bugcomment-box.pt	2010-03-29 12:29:37 +0000
@@ -74,16 +74,19 @@
       </li>
     </ul>
 
-    <div class="bug-comment"
-      tal:content="structure
-        comment/text_for_display/fmt:obfuscate-email/fmt:email-to-html">
-      Comment text.
-    </div>
-    <p tal:condition="comment/was_truncated">
-      <a tal:attributes="href comment/fmt:url">Read more...</a>
-    </p>
+    <tal:cache tal:replace="cache:authenticated,30 days,comment/index">
+      <div class="bug-comment"
+        tal:content="structure
+          comment/text_for_display/fmt:obfuscate-email/fmt:email-to-html">
+        Comment text.
+      </div>
+      <p tal:condition="comment/was_truncated">
+        <a tal:attributes="href comment/fmt:url">Read more...</a>
+      </p>
+    </tal:cache>
   </div>
-  <div class="boardCommentFooter" tal:condition="comment/show_footer">
+  <div class="boardCommentFooter" tal:condition="comment/show_footer"
+    tal:content="cache:authenticated,30 minutes,comment/index">
     <tal:activity
         define="activity_list comment/activity"
         condition="comment/activity">

=== modified file 'lib/lp/services/memcache/doc/tales-cache.txt'
--- lib/lp/services/memcache/doc/tales-cache.txt	2010-03-03 11:00:42 +0000
+++ lib/lp/services/memcache/doc/tales-cache.txt	2010-03-31 06:39:03 +0000
@@ -210,5 +210,84 @@
         <div> first 1 </div>
     </body>
 
-
+Note that the default syntax isn't useful when you are looping over
+a changeable list. For example, let's say we have a list of comments.
+The basic syntax lets us cache the first comment, the second comment,
+the third comment etc.
+
+    >>> comments = ["Today's comment", "Yesterday's comment"]
+    >>> template = TestPageTemplate(dedent("""\
+    ...     <body>
+    ...         <div tal:omit-tag="" tal:repeat="comment comments">
+    ...             <div tal:replace="cache:public">
+    ...                 <span tal:replace="comment" />
+    ...             </div>
+    ...         </div>
+    ...     </body>"""))
+    >>> print template(comments=comments)
+    <body>
+    Today's comment
+    Yesterday's comment
+    </body>
+
+However, if we now insert a new comment at the front of the list, the
+first comment has now become the second comment and our caches are
+incorrect.
+
+    >>> updated_comments = ['New comment'] + comments
+    >>> print template(comments=updated_comments)
+    <body>
+    Today's comment
+    Yesterday's comment
+    Yesterday's comment
+    </body>
+
+To correct this, we need to manually specify part of the key using
+a tales path expression. This path expression can be anything
+that can be cast to a string.
+
+    >>> template = TestPageTemplate(dedent("""\
+    ...     <body>
+    ...         <div tal:omit-tag="" tal:repeat="comment comments">
+    ...             <div tal:replace="cache:
+    ...                 public,1 hour,comment/comment_num">
+    ...                 <span tal:replace="comment/comment" />
+    ...             </div>
+    ...         </div>
+    ...     </body>"""))
+    >>> comments = [
+    ...     {'comment_num': 42, 'comment': "Today's comment"},
+    ...     {'comment_num': 12, 'comment': "Yesterday's comment"}]
+    >>> print template(comments=comments)
+    <body>
+    Today's comment
+    Yesterday's comment
+    </body>
+
+    >>> comments.insert(0, {'comment_num': 'foo', 'comment': 'New comment'})
+    >>> print template(comments=comments)
+    <body>
+    New comment
+    Today's comment
+    Yesterday's comment
+    </body>
+
+If we mutate the comments list without flushing the cache, we can
+demonstrate that the cached values are being correctly used in the loop.
+
+    >>> comments[2]['comment'] = 'Mutated comment'
+    >>> print template(comments=comments)
+    <body>
+    New comment
+    Today's comment
+    Yesterday's comment
+    </body>
+
+    >>> MemcachedLayer.purge()
+    >>> print template(comments=comments)
+    <body>
+    New comment
+    Today's comment
+    Mutated comment
+    </body>
 

=== modified file 'lib/lp/services/memcache/tales.py'
--- lib/lp/services/memcache/tales.py	2010-03-17 07:28:04 +0000
+++ lib/lp/services/memcache/tales.py	2010-03-31 06:39:03 +0000
@@ -4,7 +4,11 @@
 """Implementation of the cache: namespace in TALES."""
 
 __metaclass__ = type
-__all__ = []
+__all__ = [
+    'MemcacheExpr',
+    'MemcacheHit',
+    'MemcacheMiss',
+    ]
 
 
 from hashlib import md5
@@ -15,6 +19,7 @@
 from zope.interface import implements
 from zope.tal.talinterpreter import TALInterpreter, I18nMessageTypes
 from zope.tales.interfaces import ITALESExpression
+from zope.tales.expressions import simpleTraverse, PathExpr
 
 from canonical.base import base
 from canonical.config import config
@@ -23,6 +28,10 @@
 from lp.services.memcache.interfaces import IMemcacheClient
 
 
+# Request annotation key.
+COUNTER_KEY = 'lp.services.memcache.tales.counter'
+
+
 class MemcacheExpr:
     """Namespace to provide memcache caching of page template chunks.
 
@@ -34,7 +43,7 @@
     </div>
     """
     implements(ITALESExpression)
-    def __init__(self, name, expr, engine):
+    def __init__(self, name, expr, engine, traverser=simpleTraverse):
         """expr is in the format "visibility, 42 units".
 
         visibility is one of...
@@ -67,18 +76,30 @@
         """
         self._s = expr
 
-        if ',' in expr:
-            try:
-                self.visibility, max_age = (s.strip() for s in expr.split(','))
-            except ValueError:
-                raise SyntaxError("Too many arguments in cache: expression")
-        else:
-            self.visibility = expr.strip()
+        components = [component.strip() for component in expr.split(',')]
+        num_components = len(components)
+        if num_components == 1:
+            self.visibility = components[0]
             max_age = None
-        assert self.visibility in (
-            'anonymous', 'public', 'private', 'authenticated',
-            ), 'visibility must be anonymous, public, private or authenticated'
-
+            self.extra_key = None
+        elif num_components == 2:
+            self.visibility, max_age = components
+            self.extra_key = None
+        elif num_components == 3:
+            self.visibility, max_age, extra_key = components
+            # Construct a callable that will evaluate the subpath
+            # expression when passed a context.
+            self.extra_key = PathExpr(name, extra_key, engine, traverser)
+        else:
+            raise SyntaxError("Too many arguments in cache: expression")
+
+        if self.visibility not in (
+            'anonymous', 'public', 'private', 'authenticated'):
+            raise SyntaxError(
+                'visibility must be anonymous, public, private or '
+                'authenticated')
+
+        # Convert the max_age string to an integer number of seconds.
         if max_age is None:
             self.max_age = 0
         else:
@@ -156,21 +177,27 @@
             else: # private visibility
                 uid = str(logged_in_user.id)
 
-        # We include a counter in the key, reset at the start of the
-        # request, to ensure we get unique but repeatable keys inside
-        # tal:repeat loops.
-        counter_key = 'lp.services.memcache.tales.counter'
-        counter = request.annotations.get(counter_key, 0) + 1
-        request.annotations[counter_key] = counter
+        # The extra_key is used to differentiate items inside loops.
+        if self.extra_key is not None:
+            # Encode it to base64 to make it memcached key safe.
+            extra_key = unicode(
+                self.extra_key(econtext)).encode('base64').rstrip()
+        else:
+            # If no extra_key was specified, we include a counter in the
+            # key that is reset at the start of the request. This
+            # ensures we get unique but repeatable keys inside
+            # tal:repeat loops.
+            extra_key = request.annotations.get(COUNTER_KEY, 0) + 1
+            request.annotations[COUNTER_KEY] = extra_key
 
         # We use pt: as a unique prefix to ensure no clashes with other
         # components using the memcached servers. The order of components
         # below only matters for human readability and memcached reporting
         # tools - it doesn't really matter provided all the components are
         # included and separators used.
-        key = "pt:%s:%s,%s:%s:%d,%d:%d,%s" % (
+        key = "pt:%s:%s,%s:%s:%d,%d:%s,%s" % (
             self._lpconfig, source_file, versioninfo.revno, uid,
-            econtext.position[0], econtext.position[1], counter,
+            econtext.position[0], econtext.position[1], extra_key,
             sanitized_url,
             )
 
@@ -184,11 +211,9 @@
         return key
 
     def __call__(self, econtext):
-
         # If we have an 'anonymous' visibility chunk and are logged in,
         # we don't cache. Return the 'default' magic token to interpret
         # the contents.
-        request = econtext.getValue('request')
         if (self.visibility == 'anonymous'
             and getUtility(ILaunchBag).user is not None):
             return econtext.getDefault()

